/*
 * Schmid ADC Linux Kernel Modul
 * Copyright (C) 2006  Hanno Rein - University of Tübingen
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * */

#define ADC_MAJOR 	61
#define ADC_IRQ 	5
#define ADC_BASE 	0x380
#define ADC_LEN 	16
#define K_BUFSIZE 	1024
#define HISTO_SIZE 	16384
#define BUSY_TEST_INTERVAL HZ/100
#define TRUE 		1
#define FALSE 		0

#ifdef 	MODULE
#include <linux/module.h>
#include <linux/version.h>
MODULE_AUTHOR("Hanno Rein <hanno@hanno-rein.de>");
MODULE_DESCRIPTION("Schmid ISA-ADC Driver: irq 5, io 0x380");
MODULE_LICENSE("GPL");
#endif

#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>

#include <asm/uaccess.h>	// copy_to_user
#define memcpy_tofs copy_to_user
#include <asm/io.h>
#include "adc_lib.h"
#include "adc_lib.c"

void clear_histo(void);
void busy_test(unsigned long data);
void stop_adc(unsigned long data);

static int io[1]  = { ADC_BASE };
static int irq[1] = { ADC_IRQ  };
static unsigned int *histo;
static struct timer_list busy_timer;
static struct timer_list live_timer;
unsigned long livetime  = 0;
unsigned long deadtime1 = 0;
unsigned long deadtime2 = 0;
unsigned long expires_jiffies;
unsigned int  messung = FALSE;


// Daten auslesen
ssize_t adc_read(struct file *file, char __user * buf, size_t  count, loff_t * l) 
{
  if ( count > HISTO_SIZE * sizeof(*histo) ){
     	count = HISTO_SIZE * sizeof(*histo);
  }
  if(copy_to_user((int*)buf, histo, count)){
#ifdef DEBUG	  
	printk ("ADC: read nicht erfolgreich");
#endif	
	return -EFAULT;
  }
  return count;
}

int adc_ioctl(struct inode *inode, struct file *file,
	  unsigned int cmd, unsigned long arg)
{
  switch ( cmd ) {
  case ADC_START:
    messung = TRUE;
    irqhalf();

    busy_timer.expires = jiffies + BUSY_TEST_INTERVAL;
    add_timer(&busy_timer);

    if ( livetime > 0 ) {
#ifdef DEBUG	  
	printk("ADC: livetime=%li\n",livetime);
#endif
      	expires_jiffies = jiffies + HZ * livetime;
      	live_timer.expires = expires_jiffies;
      	add_timer(&live_timer);
    }

    run();
    break;

  case ADC_STOP:
#ifdef DEBUG	  
    printk ("ADC: STOP (jiffies: %ld)\n", jiffies);
#endif
    if ( messung ) {
      stop_adc ( 0 );
    }
    break;

  case ADC_DISK_ON:
#ifdef DEBUG	  
    printk ("ADC: DISK_ON\n");
#endif
    disk();
    break;

  case ADC_DISK_OFF:
#ifdef DEBUG	  
    printk ("ADC: DISK_OFF\n");
#endif
    disk_off();
    break;

  case ADC_SETV_DISK_LOW:
#ifdef DEBUG	  
    printk ("ADC: SETV_DISK_LOW: %ld\n", arg);
#endif
    dac(1,arg,TRUE);
    break;

  case ADC_SETV_DISK_HIGH:
#ifdef DEBUG	  
    printk ("ADC: SETV_DISK_HIGH: %ld\n", arg);
#endif
    dac(2,arg,TRUE);
    break;

  case ADC_SETV_RESET:
#ifdef DEBUG	  
    printk ("ADC: SETV_RESET: %ld\n", arg);
#endif
    dac(3,arg,TRUE);
    break;

  case ADC_SETV_ZERO:
#ifdef DEBUG	  
    printk ("ADC: SETV_ZERO: %ld\n", arg);
#endif	  
    dac(4,arg,TRUE);
    break;

  case ADC_ALLRESET:
#ifdef DEBUG	  
    printk ("ADC: ALLRESET\n");
#endif
    allreset();
    allres_release();
    fiforeset();
    clear_histo();
    break;

  case ADC_ROUTE_ON:
#ifdef DEBUG	  
    printk ("ADC: ROUTE_ON\n");
#endif
    route();
    break;

  case ADC_ROUTE_OFF:
#ifdef DEBUG	  
    printk ("ADC: ROUTE_OFF\n");
#endif
    route_off();
    break;

  case ADC_GATE_ON:
#ifdef DEBUG	  
    printk ("ADC: GATE_ON\n");
#endif
    extgate();
    break;

  case ADC_GATE_OFF:
#ifdef DEBUG	  
    printk ("ADC: GATE_OFF\n");
#endif
    extgate_off();
    break;

  case ADC_TIMER_SET:
    if ( messung == FALSE ) {
      livetime = arg;
    } else {
      return -1;
    }
    break;

  case ADC_DEAD_TIME1:
    put_user(deadtime1,(unsigned long *) arg);
    break;
  
  case ADC_DEAD_TIME2:
    put_user(deadtime2,(unsigned long *) arg);
    break;

  case ADC_TIMER_STATUS:
    if ( messung && livetime > 0 ) {
	put_user(expires_jiffies - jiffies, (unsigned long *) arg);
	
    } else {
	put_user(0, (unsigned long *) arg);
    }
    break;
    
  }
  return 0;
}

ssize_t adc_write(struct file *filp, const char __user *buf, size_t count, loff_t *f_pos)
{
  return count;
}


static int adc_open(struct inode *inode, struct file *file)
{
#ifdef DEBUG	  
  printk ("ADC: open\n");
#endif
  allreset ();
  allres_release ();
  fiforeset ();
  disk_off ();
  extgate_off();
  route_off();
  dac (1,0,FALSE);
  dac (2,4093,FALSE);
  dac (3,1500,FALSE);
  dac (4,0,TRUE);
  clear_histo();

  return 0;
}


int adc_release(struct inode *inode, struct file *filp)
{
#ifdef DEBUG	  
  printk ("ADC: close\n");
#endif
  if ( messung ) stop_adc( 0 );
  return 0;
}

static struct file_operations adc_fops = {
 .owner 	=  THIS_MODULE,
 .read 		=  adc_read,
 .write 	=  adc_write,
 .ioctl 	=  adc_ioctl,  
 .open 		=  adc_open,
 .release 	=  adc_release,
};

static irqreturn_t adc_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
  int i;
  for ( i = 0; (i < 256)&&(status()&0x2); i++ )  (*(histo+adc()))++;
  return IRQ_HANDLED;
}

int init_module(void)
{
  int num;

#ifdef DEBUG	  
  printk ("ADC: io requested: %x   irq: %d\n", io[0],  irq[0]); 
#endif
  if ( register_chrdev(ADC_MAJOR, "adc_karte", &adc_fops) == -EBUSY) {
    printk("ADC: unable to get major for adc device\n");
    return -EIO;
  }

  if ( request_region ( io[0], ADC_LEN, "adc_karte" )==NULL ) {
    printk("ADC: unable to access i/o ports for adc device\n");
    unregister_chrdev(ADC_MAJOR, "adc_karte");
    return -EIO;
  }

  adc_setbase(io[0]);

  if ( request_irq ( irq[0], adc_interrupt, SA_INTERRUPT, "adc_karte", NULL ) ) {
    printk("ADC: unable to install irq %d for adc device\n", ADC_IRQ);
    release_region ( io[0], ADC_LEN );
    unregister_chrdev(ADC_MAJOR, "adc_karte");
    return -EIO;
  }

  if ( ! (histo = kmalloc ( HISTO_SIZE * sizeof(*histo), GFP_KERNEL )) ) {
    printk ("ADC: unable to allocate %d bytes of kernel memory for adc device\n", 
	    HISTO_SIZE * sizeof(*histo));

    free_irq ( irq[0] , NULL);
    release_region ( io[0], ADC_LEN );
    unregister_chrdev(ADC_MAJOR, "adc_karte");
    return -EIO;
  } else {
#ifdef DEBUG	  
    printk ("ADC: kmalloc ok: %d Bytes\n", HISTO_SIZE * sizeof(*histo));
#endif
    clear_histo ();
  }
  num = readnum();
  if ( num < 0 || num > 200 ) {
    printk ("ADC: no card found at io=0x%x (num=0x%x)\n", io[0], num);
    free_irq ( irq[0] , NULL);
    release_region ( io[0], ADC_LEN );
    unregister_chrdev(ADC_MAJOR, "adc_karte");
    kfree ( histo );
    return -EIO;
  }

  printk("ADC: driver installed: major: %d io_base: 0x%x irq: %d card no: %d\n", 
	 ADC_MAJOR, io[0], irq[0], num);

  init_timer(&busy_timer);
  busy_timer.function = busy_test;
  init_timer(&live_timer);
  live_timer.function = stop_adc;

  return 0;
}

void cleanup_module(void)
{
  printk("ADC: driver removed: major: %d io_base: 0x%x irq: %d\n", 
	 ADC_MAJOR, ADC_BASE, ADC_IRQ);
  free_irq ( ADC_IRQ , NULL);
  release_region ( ADC_BASE, ADC_LEN );
  unregister_chrdev(ADC_MAJOR, "adc_karte");
  kfree ( histo );
}


void clear_histo(void)
{
  int i;
  deadtime1=0;
  deadtime2=0;
  for ( i = 0; i < HISTO_SIZE; i++ )  *(histo + i) = 0;
}

void busy_test(unsigned long data)
{
  del_timer(&busy_timer);
  if (status() & 1){
	deadtime1++;
  }else{
	deadtime2++;
  }
  busy_timer.expires = jiffies + BUSY_TEST_INTERVAL;
  add_timer(&busy_timer);
}

void stop_adc(unsigned long data)
{
  int t=0;
#ifdef DEBUG	  
  printk ("ADC: Messung wird gestoppt\n");
#endif
  messung = FALSE;
  
  del_timer(&busy_timer);
#ifdef DEBUG	  
  printk ("ADC: Totzeit %lu \t %lu \n",deadtime1,deadtime2);
#endif
  if ( livetime >0 ) del_timer(&live_timer);
  stop();

  while ( (status() & 0x2) && (t<10000) ) {
    	(*(histo+adc()))++;
	t++;
  }
}

